package com.ho.test.aop.cache.service.common;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.management.ObjectInstance;
import javax.management.ObjectName;

import org.aopalliance.intercept.MethodInvocation;

import com.ho.test.mbean.common.MbeanTools;

/**
 * <p>Title: CacheInterceptor</p>
 * <p>Copyright: Copyright (c) 2010</p>
 * <p>Company: HUO</p>
 * 
 * <p>
 * MethodInterceptor that looks for the requested method output in the cache and returns it if found.
 * <br/>If not found, it execute the requested method, cache the output and returns it. 
 * </p>
 * 
 * <p>
 * This assumes that the toString().toLowerCase() on each the arguments of the intercepted methods 
 * can be considered unique as they are used to to build the keys that will be used to store 
 * the Objects in the cache.
 * </p>
 * 
 * @author Huseyin OZVEREN
 */
public class CacheInterceptor implements ICacheInterceptor {

	// -------------------------------------------------------------- CONSTANTS
	// ----------------------------------------------------- PRIVATE ATTRIBUTES

	/**
	 * Contains all the methods name to cache.
	 */
	private List<String> methodsToCache = new ArrayList<String>();

	/**
	 * Contains all the cached outputs.
	 * <br/>The items stored as Object are stored with a key in the cache corresponding to :
	 * <br/>method name + parameters separated with "_";
	 */
	private Cache<String, Object> cache = new Cache<String, Object>();

	/**
	 * Time in seconds that returned object will stay active in the cache.
	 */
	private long cacheTimeInSeconds = 0;

	// ------------------------------------------------------------- EXCEPTIONS
	// ------------------------------------------------------- INTERNAL CLASSES
	// ----------------------------------------------------------- CONSTRUCTORS
	// --------------------------------------------------------- PUBLIC METHODS

	/**
	 * Method invoke
	 * @param arg0
	 * @return
	 * @throws Throwable
	 * @see org.aopalliance.intercept.MethodInterceptor#invoke(org.aopalliance.intercept.MethodInvocation)
	 */
	public Object invoke(MethodInvocation method) throws Throwable {

		if (-1 != methodsToCache.indexOf(method.getMethod().getName())) {

			// Computes the cache key
			String cacheKey = getCacheKey(method);

			// Try to retrieve the cache output from the cache
			Object output = null;
			{
				output = cache.get(cacheKey, cacheTimeInSeconds);
			} // end-block

			// If the output is already present in the cache
			// return it
			if (null != output) {
				return output;
			} // end-if

			// Else, compute the output,
			// store it in the cache
			// and return it.
			output = method.proceed();

			cache.put(cacheKey, output);

			return output;
		} else {
			Object output = null;
			output = method.proceed();
			return output;

		} // end-if
	}

	/**
	 * Erase all the items stored in the cache.
	 */
	public void clear() {
		cache.clear();
	}

	// -------------------------------------------------------- PRIVATE METHODS


	/**
	 * Get the cache key
	 * @return the key used to store the items in the cache.
	 */
	private String getCacheKey(MethodInvocation method) {
	    StringBuffer cacheNameBuffer = new StringBuffer();
	    cacheNameBuffer.append(method.getMethod().getDeclaringClass().getName() + "." + method.getMethod().getName() + "(");
		//cacheNameBuffer.append(method.getMethod().getName());

		if (method.getArguments() != null && method.getArguments().length > 0) {
			for (int a = 0; a < method.getArguments().length; a++) {
				cacheNameBuffer.append("-");
				if (method.getArguments()[a] == null) {
					cacheNameBuffer.append("null");
				} else {
					cacheNameBuffer.append(method.getArguments()[a].toString()
							.toLowerCase());
				} // end-if
			} // end-for
		} // end-if

		return cacheNameBuffer.toString();
	}
	
	
	// ------------------------------------------------------ PROTECTED METHODS
	// --------------------------------------------- ADVANCED GETTERS & SETTERS
	// ------------------------------------------------------ GETTERS & SETTERS

	public final List<String> getMethodsToCache() {
		return this.methodsToCache;
	}

	public final void setMethodsToCache(List<String> methodsToCache) {
		this.methodsToCache = methodsToCache;
	}

	public final long getCacheTimeInSeconds() {
		return this.cacheTimeInSeconds;
	}

	public final void setCacheTimeInSeconds(long cacheTimeInSeconds) {
		this.cacheTimeInSeconds = cacheTimeInSeconds;
	}

	public final String getCacheKeysCore4MBEANTestOnly(){
		if(cache!=null){
			return cache.getCacheKeysCore4MBEANTestOnly();
		}
		return null;
	}
	
	public Map<ObjectName, ObjectInstance> getMbeanByPattern4MBEANTestOnly(String mbeanPattern){
		try {
			return new MbeanTools().getMbeanByPattern(mbeanPattern);
		} catch (Throwable th) {
			th.printStackTrace();
		}
		
		return null;
	}
}